[vinagre] Added an animation in the tab, giving feedback on slow connections.



commit fd8b0c423590fdd11852b4e5c2126a36c99d54c0
Author: Jonh Wendell <jwendell gnome org>
Date:   Thu Aug 6 15:22:37 2009 -0300

    Added an animation in the tab, giving feedback on slow connections.
    
    Closes #580674

 vinagre/Makefile.am        |    3 +
 vinagre/vinagre-notebook.c |   35 ++-
 vinagre/vinagre-spinner.c  |  983 ++++++++++++++++++++++++++++++++++++++++++++
 vinagre/vinagre-spinner.h  |   89 ++++
 4 files changed, 1106 insertions(+), 4 deletions(-)
---
diff --git a/vinagre/Makefile.am b/vinagre/Makefile.am
index 4a4beb9..1022b45 100644
--- a/vinagre/Makefile.am
+++ b/vinagre/Makefile.am
@@ -25,6 +25,7 @@ NOINST_H_FILES = \
   vinagre-plugin-dialog.h \
   vinagre-plugins-engine.h \
   vinagre-window-private.h \
+  vinagre-spinner.h \
   $(NULL)
 
 INST_H_FILES = \
@@ -89,6 +90,7 @@ libvinagre_la_SOURCES = \
   vinagre-tab.c \
   vinagre-utils.c \
   vinagre-window.c \
+  vinagre-spinner.c \
   $(NOINST_H_FILES) \
   $(INST_H_FILES) \
   $(NULL)
@@ -178,6 +180,7 @@ vinagre_applet_SOURCES =					\
 	vinagre-dirs.h vinagre-dirs.c \
 	vinagre-fav.h vinagre-fav.c \
 	vinagre-connect.h vinagre-connect.c \
+	vinagre-spinner.h vinagre-spinner.c \
 	$(NULL)
 
 if AVAHI
diff --git a/vinagre/vinagre-spinner.c b/vinagre/vinagre-spinner.c
new file mode 100644
index 0000000..1cb8009
--- /dev/null
+++ b/vinagre/vinagre-spinner.c
@@ -0,0 +1,983 @@
+/*
+ * vinagre-spinner.c
+ * This file is part of vinagre
+ *
+ * Copyright (C) 2005 - Paolo Maggi 
+ * Copyright (C) 2002-2004 Marco Pesenti Gritti
+ * Copyright (C) 2004 Christian Persch
+ * Copyright (C) 2000 - Eazel, Inc. 
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, 
+ * Boston, MA 02111-1307, USA.
+ */
+ 
+/*
+ * This widget was originally written by Andy Hertzfeld <andy eazel com> for
+ * Nautilus. It was then modified by Marco Pesenti Gritti and Christian Persch
+ * for Epiphany.
+ */
+
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
+#include "vinagre-spinner.h"
+
+#include <gdk-pixbuf/gdk-pixbuf.h>
+#include <gtk/gtk.h>
+
+/* Spinner cache implementation */
+
+#define VINAGRE_TYPE_SPINNER_CACHE		(vinagre_spinner_cache_get_type())
+#define VINAGRE_SPINNER_CACHE(object)		(G_TYPE_CHECK_INSTANCE_CAST((object), VINAGRE_TYPE_SPINNER_CACHE, VinagreSpinnerCache))
+#define VINAGRE_SPINNER_CACHE_CLASS(klass) 	(G_TYPE_CHECK_CLASS_CAST((klass), VINAGRE_TYPE_SPINNER_CACHE, VinagreSpinnerCacheClass))
+#define VINAGRE_IS_SPINNER_CACHE(object)		(G_TYPE_CHECK_INSTANCE_TYPE((object), VINAGRE_TYPE_SPINNER_CACHE))
+#define VINAGRE_IS_SPINNER_CACHE_CLASS(klass)	(G_TYPE_CHECK_CLASS_TYPE((klass), VINAGRE_TYPE_SPINNER_CACHE))
+#define VINAGRE_SPINNER_CACHE_GET_CLASS(obj)	(G_TYPE_INSTANCE_GET_CLASS((obj), VINAGRE_TYPE_SPINNER_CACHE, VinagreSpinnerCacheClass))
+
+typedef struct _VinagreSpinnerCache		VinagreSpinnerCache;
+typedef struct _VinagreSpinnerCacheClass		VinagreSpinnerCacheClass;
+typedef struct _VinagreSpinnerCachePrivate	VinagreSpinnerCachePrivate;
+
+struct _VinagreSpinnerCacheClass
+{
+	GObjectClass parent_class;
+};
+
+struct _VinagreSpinnerCache
+{
+	GObject parent_object;
+
+	/*< private >*/
+	VinagreSpinnerCachePrivate *priv;
+};
+
+#define VINAGRE_SPINNER_CACHE_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), VINAGRE_TYPE_SPINNER_CACHE, VinagreSpinnerCachePrivate))
+
+struct _VinagreSpinnerCachePrivate
+{
+	/* Hash table of GdkScreen -> VinagreSpinnerCacheData */
+	GHashTable *hash;
+};
+
+typedef struct
+{
+	guint        ref_count;
+	GtkIconSize  size;
+	gint         width;
+	gint         height;
+	GdkPixbuf  **animation_pixbufs;
+	guint        n_animation_pixbufs;
+} VinagreSpinnerImages;
+
+#define LAST_ICON_SIZE			GTK_ICON_SIZE_DIALOG + 1
+#define SPINNER_ICON_NAME		"process-working"
+#define SPINNER_FALLBACK_ICON_NAME	"gnome-spinner"
+#define VINAGRE_SPINNER_IMAGES_INVALID	((VinagreSpinnerImages *) 0x1)
+
+typedef struct
+{
+	GdkScreen          *screen;
+	GtkIconTheme       *icon_theme;
+	VinagreSpinnerImages *images[LAST_ICON_SIZE];
+} VinagreSpinnerCacheData;
+
+static void vinagre_spinner_cache_class_init	(VinagreSpinnerCacheClass *klass);
+static void vinagre_spinner_cache_init		(VinagreSpinnerCache      *cache);
+
+static GObjectClass *vinagre_spinner_cache_parent_class;
+
+static GType
+vinagre_spinner_cache_get_type (void)
+{
+	static GType type = 0;
+
+	if (G_UNLIKELY (type == 0))
+	{
+		const GTypeInfo our_info =
+		{
+			sizeof (VinagreSpinnerCacheClass),
+			NULL,
+			NULL,
+			(GClassInitFunc) vinagre_spinner_cache_class_init,
+			NULL,
+			NULL,
+			sizeof (VinagreSpinnerCache),
+			0,
+			(GInstanceInitFunc) vinagre_spinner_cache_init
+		};
+
+		type = g_type_register_static (G_TYPE_OBJECT,
+					       "VinagreSpinnerCache",
+					       &our_info, 0);
+	}
+
+	return type;
+}
+
+static VinagreSpinnerImages *
+vinagre_spinner_images_ref (VinagreSpinnerImages *images)
+{
+	g_return_val_if_fail (images != NULL, NULL);
+
+	images->ref_count++;
+
+	return images;
+}
+
+static void
+vinagre_spinner_images_unref (VinagreSpinnerImages *images)
+{
+	g_return_if_fail (images != NULL);
+
+	images->ref_count--;
+	if (images->ref_count == 0)
+	{
+		guint i;
+
+		/* LOG ("Freeing spinner images %p for size %d", images, images->size); */
+
+		for (i = 0; i < images->n_animation_pixbufs; ++i)
+		{
+			g_object_unref (images->animation_pixbufs[i]);
+		}
+		g_free (images->animation_pixbufs);
+
+		g_free (images);
+	}
+}
+
+static void
+vinagre_spinner_cache_data_unload (VinagreSpinnerCacheData *data)
+{
+	GtkIconSize size;
+	VinagreSpinnerImages *images;
+
+	g_return_if_fail (data != NULL);
+
+	/* LOG ("VinagreSpinnerDataCache unload for screen %p", data->screen); */
+
+	for (size = GTK_ICON_SIZE_INVALID; size < LAST_ICON_SIZE; ++size)
+	{
+		images = data->images[size];
+		data->images[size] = NULL;
+
+		if (images != NULL && images != VINAGRE_SPINNER_IMAGES_INVALID)
+		{
+			vinagre_spinner_images_unref (images);
+		}
+	}
+}
+
+static GdkPixbuf *
+extract_frame (GdkPixbuf *grid_pixbuf,
+	       int x,
+	       int y,
+	       int size)
+{
+	GdkPixbuf *pixbuf;
+
+	if (x + size > gdk_pixbuf_get_width (grid_pixbuf) ||
+	    y + size > gdk_pixbuf_get_height (grid_pixbuf))
+	{
+		return NULL;
+	}
+
+	pixbuf = gdk_pixbuf_new_subpixbuf (grid_pixbuf,
+					   x, y,
+					   size, size);
+	g_return_val_if_fail (pixbuf != NULL, NULL);
+
+	return pixbuf;
+}
+
+static GdkPixbuf *
+scale_to_size (GdkPixbuf *pixbuf,
+	       int dw,
+	       int dh)
+{
+	GdkPixbuf *result;
+	int pw, ph;
+
+	g_return_val_if_fail (pixbuf != NULL, NULL);
+
+	pw = gdk_pixbuf_get_width (pixbuf);
+	ph = gdk_pixbuf_get_height (pixbuf);
+
+	if (pw != dw || ph != dh)
+	{
+		result = gdk_pixbuf_scale_simple (pixbuf, dw, dh,
+						  GDK_INTERP_BILINEAR);
+		g_object_unref (pixbuf);
+		return result;
+	}
+
+	return pixbuf;
+}
+
+static VinagreSpinnerImages *
+vinagre_spinner_images_load (GdkScreen *screen,
+			   GtkIconTheme *icon_theme,
+			   GtkIconSize icon_size)
+{
+	VinagreSpinnerImages *images;
+	GdkPixbuf *icon_pixbuf, *pixbuf;
+	GtkIconInfo *icon_info = NULL;
+	int grid_width, grid_height, x, y, requested_size, size, isw, ish, n;
+	const char *icon;
+	GSList *list = NULL, *l;
+
+	/* LOG ("VinagreSpinnerCacheData loading for screen %p at size %d", screen, icon_size); */
+
+	/* START_PROFILER ("loading spinner animation") */
+
+	if (!gtk_icon_size_lookup_for_settings (gtk_settings_get_for_screen (screen),
+						icon_size, &isw, &ish))
+		goto loser;
+ 
+	requested_size = MAX (ish, isw);
+
+	/* Load the animation. The 'rest icon' is the 0th frame */
+	icon_info = gtk_icon_theme_lookup_icon (icon_theme,
+						SPINNER_ICON_NAME,
+						requested_size, 0);
+	if (icon_info == NULL)
+	{
+		g_warning ("Throbber animation not found");
+
+		/* If the icon naming spec compliant name wasn't found, try the old name */
+		icon_info = gtk_icon_theme_lookup_icon (icon_theme,
+							SPINNER_FALLBACK_ICON_NAME,
+						        requested_size, 0);
+		if (icon_info == NULL)
+		{
+			g_warning ("Throbber fallback animation not found either");
+			goto loser;
+	 	}
+	}
+
+	g_assert (icon_info != NULL);
+
+	size = gtk_icon_info_get_base_size (icon_info);
+	icon = gtk_icon_info_get_filename (icon_info);
+
+	if (icon == NULL)
+		goto loser;
+
+	icon_pixbuf = gdk_pixbuf_new_from_file (icon, NULL);
+	gtk_icon_info_free (icon_info);
+	icon_info = NULL;
+
+	if (icon_pixbuf == NULL)
+	{
+		g_warning ("Could not load the spinner file");
+		goto loser;
+	}
+
+	grid_width = gdk_pixbuf_get_width (icon_pixbuf);
+	grid_height = gdk_pixbuf_get_height (icon_pixbuf);
+
+	n = 0;
+	for (y = 0; y < grid_height; y += size)
+	{
+		for (x = 0; x < grid_width ; x += size)
+		{
+			pixbuf = extract_frame (icon_pixbuf, x, y, size);
+
+			if (pixbuf)
+			{
+				list = g_slist_prepend (list, pixbuf);
+				++n;
+			}
+			else
+			{
+				g_warning ("Cannot extract frame (%d, %d) from the grid\n", x, y);
+			}
+		}
+	}
+
+	g_object_unref (icon_pixbuf);
+
+	if (list == NULL)
+		goto loser;
+
+	/* g_assert (n > 0); */
+
+	if (size > requested_size)
+	{
+		for (l = list; l != NULL; l = l->next)
+		{
+			l->data = scale_to_size (l->data, isw, ish);
+		}
+ 	}
+
+	/* Now we've successfully got all the data */
+	images = g_new (VinagreSpinnerImages, 1);
+	images->ref_count = 1;
+ 
+	images->size = icon_size;
+	images->width = images->height = requested_size;
+
+	images->n_animation_pixbufs = n;
+	images->animation_pixbufs = g_new (GdkPixbuf *, n);
+
+	for (l = list; l != NULL; l = l->next)
+	{
+		g_assert (l->data != NULL);
+		images->animation_pixbufs[--n] = l->data;
+	}
+	g_assert (n == 0);
+
+	g_slist_free (list);
+
+	/* STOP_PROFILER ("loading spinner animation") */
+	return images;
+ 
+loser:
+	if (icon_info)
+	{
+		gtk_icon_info_free (icon_info);
+ 	}
+
+	g_slist_foreach (list, (GFunc) g_object_unref, NULL);
+
+	/* STOP_PROFILER ("loading spinner animation") */
+
+	return NULL;
+}
+
+static VinagreSpinnerCacheData *
+vinagre_spinner_cache_data_new (GdkScreen *screen)
+{
+	VinagreSpinnerCacheData *data;
+
+	data = g_new0 (VinagreSpinnerCacheData, 1);
+
+	data->screen = screen;
+	data->icon_theme = gtk_icon_theme_get_for_screen (screen);
+	g_signal_connect_swapped (data->icon_theme,
+				  "changed",
+				  G_CALLBACK (vinagre_spinner_cache_data_unload),
+				  data);
+
+	return data;
+}
+
+static void
+vinagre_spinner_cache_data_free (VinagreSpinnerCacheData *data)
+{
+	g_return_if_fail (data != NULL);
+	g_return_if_fail (data->icon_theme != NULL);
+
+	g_signal_handlers_disconnect_by_func (data->icon_theme,
+					      G_CALLBACK (vinagre_spinner_cache_data_unload),
+					      data);
+
+	vinagre_spinner_cache_data_unload (data);
+
+	g_free (data);
+}
+
+static VinagreSpinnerImages *
+vinagre_spinner_cache_get_images (VinagreSpinnerCache *cache,
+				GdkScreen         *screen,
+				GtkIconSize        icon_size)
+{
+	VinagreSpinnerCachePrivate *priv = cache->priv;
+	VinagreSpinnerCacheData *data;
+	VinagreSpinnerImages *images;
+
+	/* LOG ("Getting animation images for screen %p at size %d", screen, icon_size); */
+
+	g_return_val_if_fail (icon_size >= 0 && icon_size < LAST_ICON_SIZE, NULL);
+
+	/* Backward compat: "invalid" meant "native" size which doesn't exist anymore */
+	if (icon_size == GTK_ICON_SIZE_INVALID)
+	{
+		icon_size = GTK_ICON_SIZE_DIALOG;
+	}
+
+	data = g_hash_table_lookup (priv->hash, screen);
+	if (data == NULL)
+	{
+		data = vinagre_spinner_cache_data_new (screen);
+		/* FIXME: think about what happens when the screen's display is closed later on */
+		g_hash_table_insert (priv->hash, screen, data);
+	}
+
+	images = data->images[icon_size];
+	if (images == VINAGRE_SPINNER_IMAGES_INVALID)
+	{
+		/* Load failed, but don't try endlessly again! */
+		return NULL;
+	}
+
+	if (images != NULL)
+	{
+		/* Return cached data */
+		return vinagre_spinner_images_ref (images);
+	}
+
+	images = vinagre_spinner_images_load (screen, data->icon_theme, icon_size);
+
+	if (images == NULL)
+ 	{
+		/* Mark as failed-to-load */
+		data->images[icon_size] = VINAGRE_SPINNER_IMAGES_INVALID;
+ 
+		return NULL;
+ 	}
+
+	data->images[icon_size] = images;
+
+	return vinagre_spinner_images_ref (images);
+}
+
+static void
+vinagre_spinner_cache_init (VinagreSpinnerCache *cache)
+{
+	VinagreSpinnerCachePrivate *priv;
+
+	priv = cache->priv = VINAGRE_SPINNER_CACHE_GET_PRIVATE (cache);
+
+	/* LOG ("VinagreSpinnerCache initialising"); */
+
+	priv->hash = g_hash_table_new_full (g_direct_hash,
+					    g_direct_equal,
+					    NULL,
+					    (GDestroyNotify) vinagre_spinner_cache_data_free);
+}
+
+static void
+vinagre_spinner_cache_finalize (GObject *object)
+{
+	VinagreSpinnerCache *cache = VINAGRE_SPINNER_CACHE (object); 
+	VinagreSpinnerCachePrivate *priv = cache->priv;
+
+	g_hash_table_destroy (priv->hash);
+
+	/* LOG ("VinagreSpinnerCache finalised"); */
+
+	G_OBJECT_CLASS (vinagre_spinner_cache_parent_class)->finalize (object);
+}
+
+static void
+vinagre_spinner_cache_class_init (VinagreSpinnerCacheClass *klass)
+{
+	GObjectClass *object_class = G_OBJECT_CLASS (klass);
+
+	vinagre_spinner_cache_parent_class = g_type_class_peek_parent (klass);
+
+	object_class->finalize = vinagre_spinner_cache_finalize;
+
+	g_type_class_add_private (object_class, sizeof (VinagreSpinnerCachePrivate));
+}
+
+static VinagreSpinnerCache *spinner_cache = NULL;
+
+static VinagreSpinnerCache *
+vinagre_spinner_cache_ref (void)
+{
+	if (spinner_cache == NULL)
+	{
+		VinagreSpinnerCache **cache_ptr;
+
+		spinner_cache = g_object_new (VINAGRE_TYPE_SPINNER_CACHE, NULL);
+		cache_ptr = &spinner_cache;
+		g_object_add_weak_pointer (G_OBJECT (spinner_cache),
+					   (gpointer *) cache_ptr);
+
+		return spinner_cache;
+	}
+
+	return g_object_ref (spinner_cache);
+}
+
+/* Spinner implementation */
+
+#define SPINNER_TIMEOUT 125 /* ms */
+
+#define VINAGRE_SPINNER_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), VINAGRE_TYPE_SPINNER, VinagreSpinnerPrivate))
+
+struct _VinagreSpinnerPrivate
+{
+	GtkIconTheme       *icon_theme;
+	VinagreSpinnerCache  *cache;
+	GtkIconSize         size;
+	VinagreSpinnerImages *images;
+	guint               current_image;
+	guint               timeout;
+	guint               timer_task;
+	guint               spinning : 1;
+	guint               need_load : 1;
+};
+
+static void vinagre_spinner_class_init	(VinagreSpinnerClass *class);
+static void vinagre_spinner_init		(VinagreSpinner      *spinner);
+
+static GObjectClass *parent_class;
+
+GType
+vinagre_spinner_get_type (void)
+{
+	static GType type = 0;
+
+	if (G_UNLIKELY (type == 0))
+	{
+		const GTypeInfo our_info =
+		{
+			sizeof (VinagreSpinnerClass),
+			NULL, /* base_init */
+			NULL, /* base_finalize */
+			(GClassInitFunc) vinagre_spinner_class_init,
+			NULL,
+			NULL, /* class_data */
+			sizeof (VinagreSpinner),
+			0, /* n_preallocs */
+			(GInstanceInitFunc) vinagre_spinner_init
+		};
+
+		type = g_type_register_static (GTK_TYPE_WIDGET,
+					       "VinagreSpinner",
+					       &our_info, 0);
+	}
+
+	return type;
+}
+
+static gboolean
+vinagre_spinner_load_images (VinagreSpinner *spinner)
+{
+	VinagreSpinnerPrivate *priv = spinner->priv;
+
+	if (priv->need_load)
+	{
+		/* START_PROFILER ("vinagre_spinner_load_images") */
+
+		priv->images =
+			vinagre_spinner_cache_get_images (priv->cache,
+							gtk_widget_get_screen (GTK_WIDGET (spinner)),
+							priv->size);
+
+		/* STOP_PROFILER ("vinagre_spinner_load_images") */
+
+		priv->current_image = 0; /* 'rest' icon */
+		priv->need_load = FALSE;
+	}
+
+	return priv->images != NULL;
+}
+
+static void
+vinagre_spinner_unload_images (VinagreSpinner *spinner)
+{
+	VinagreSpinnerPrivate *priv = spinner->priv;
+
+	if (priv->images != NULL)
+	{
+		vinagre_spinner_images_unref (priv->images);
+		priv->images = NULL;
+	}
+
+	priv->current_image = 0;
+	priv->need_load = TRUE;
+}
+
+static void
+icon_theme_changed_cb (GtkIconTheme *icon_theme,
+		       VinagreSpinner *spinner)
+{
+	vinagre_spinner_unload_images (spinner);
+	gtk_widget_queue_resize (GTK_WIDGET (spinner));
+}
+
+static void
+vinagre_spinner_init (VinagreSpinner *spinner)
+{
+	VinagreSpinnerPrivate *priv;
+
+	priv = spinner->priv = VINAGRE_SPINNER_GET_PRIVATE (spinner);
+
+	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (spinner), GTK_NO_WINDOW);
+
+	priv->cache = vinagre_spinner_cache_ref ();
+	priv->size = GTK_ICON_SIZE_DIALOG;
+	priv->spinning = FALSE;
+	priv->timeout = SPINNER_TIMEOUT;
+	priv->need_load = TRUE;
+}
+
+static int
+vinagre_spinner_expose (GtkWidget      *widget,
+		      GdkEventExpose *event)
+{
+	VinagreSpinner *spinner = VINAGRE_SPINNER (widget);
+	VinagreSpinnerPrivate *priv = spinner->priv;
+	VinagreSpinnerImages *images;
+	GdkPixbuf *pixbuf;
+	GdkGC *gc;
+	int x_offset, y_offset, width, height;
+	GdkRectangle pix_area, dest;
+
+	if (!GTK_WIDGET_DRAWABLE (spinner))
+	{
+		return FALSE;
+	}
+
+	if (priv->need_load &&
+	    !vinagre_spinner_load_images (spinner))
+	{
+		return FALSE;
+	}
+
+	images = priv->images;
+	if (images == NULL)
+	{
+		return FALSE;
+	}
+
+	/* Otherwise |images| will be NULL anyway */
+	g_assert (images->n_animation_pixbufs > 0);
+
+	g_assert (priv->current_image >= 0 &&
+		  priv->current_image < images->n_animation_pixbufs);
+
+	pixbuf = images->animation_pixbufs[priv->current_image];
+
+	g_assert (pixbuf != NULL);
+
+	width = gdk_pixbuf_get_width (pixbuf);
+	height = gdk_pixbuf_get_height (pixbuf);
+
+	/* Compute the offsets for the image centered on our allocation */
+	x_offset = (widget->allocation.width - width) / 2;
+	y_offset = (widget->allocation.height - height) / 2;
+
+	pix_area.x = x_offset + widget->allocation.x;
+	pix_area.y = y_offset + widget->allocation.y;
+	pix_area.width = width;
+	pix_area.height = height;
+
+	if (!gdk_rectangle_intersect (&event->area, &pix_area, &dest))
+	{
+		return FALSE;
+	}
+
+	gc = gdk_gc_new (widget->window);
+	gdk_draw_pixbuf (widget->window, gc, pixbuf,
+			 dest.x - x_offset - widget->allocation.x,
+			 dest.y - y_offset - widget->allocation.y,
+			 dest.x, dest.y,
+			 dest.width, dest.height,
+			 GDK_RGB_DITHER_MAX, 0, 0);
+	g_object_unref (gc);
+
+	return FALSE;
+}
+
+static gboolean
+bump_spinner_frame_cb (VinagreSpinner *spinner)
+{
+	VinagreSpinnerPrivate *priv = spinner->priv;
+
+	/* This can happen when we've unloaded the images on a theme
+	 * change, but haven't been in the queued size request yet.
+	 * Just skip this update.
+	 */
+	if (priv->images == NULL)
+		return TRUE;
+
+	priv->current_image++;
+	if (priv->current_image >= priv->images->n_animation_pixbufs)
+	{
+		/* the 0th frame is the 'rest' icon */
+		priv->current_image = MIN (1, priv->images->n_animation_pixbufs);
+	}
+
+	gtk_widget_queue_draw (GTK_WIDGET (spinner));
+
+	/* run again */
+	return TRUE;
+}
+
+/**
+ * vinagre_spinner_start:
+ * @spinner: a #VinagreSpinner
+ *
+ * Start the spinner animation.
+ **/
+void
+vinagre_spinner_start (VinagreSpinner *spinner)
+{
+	VinagreSpinnerPrivate *priv = spinner->priv;
+
+	priv->spinning = TRUE;
+
+	if (GTK_WIDGET_MAPPED (GTK_WIDGET (spinner)) &&
+			       priv->timer_task == 0 &&
+			       vinagre_spinner_load_images (spinner))
+	{
+		/* the 0th frame is the 'rest' icon */
+		priv->current_image = MIN (1, priv->images->n_animation_pixbufs);
+
+		priv->timer_task = g_timeout_add_full (G_PRIORITY_LOW,
+						       priv->timeout,
+						       (GSourceFunc) bump_spinner_frame_cb,
+						       spinner,
+						       NULL);
+	}
+}
+
+static void
+vinagre_spinner_remove_update_callback (VinagreSpinner *spinner)
+{
+	VinagreSpinnerPrivate *priv = spinner->priv;
+
+	if (priv->timer_task != 0)
+	{
+		g_source_remove (priv->timer_task);
+		priv->timer_task = 0;
+	}
+}
+
+/**
+ * vinagre_spinner_stop:
+ * @spinner: a #VinagreSpinner
+ *
+ * Stop the spinner animation.
+ **/
+void
+vinagre_spinner_stop (VinagreSpinner *spinner)
+{
+	VinagreSpinnerPrivate *priv = spinner->priv;
+
+	priv->spinning = FALSE;
+	priv->current_image = 0;
+
+	if (priv->timer_task != 0)
+	{
+		vinagre_spinner_remove_update_callback (spinner);
+
+		if (GTK_WIDGET_MAPPED (GTK_WIDGET (spinner)))
+			gtk_widget_queue_draw (GTK_WIDGET (spinner));
+	}
+}
+
+/*
+ * vinagre_spinner_set_size:
+ * @spinner: a #VinagreSpinner
+ * @size: the size of type %GtkIconSize
+ *
+ * Set the size of the spinner.
+ **/
+void
+vinagre_spinner_set_size (VinagreSpinner *spinner,
+			GtkIconSize size)
+{
+	if (size == GTK_ICON_SIZE_INVALID)
+	{
+		size = GTK_ICON_SIZE_DIALOG;
+	}
+
+	if (size != spinner->priv->size)
+	{
+		vinagre_spinner_unload_images (spinner);
+
+		spinner->priv->size = size;
+
+		gtk_widget_queue_resize (GTK_WIDGET (spinner));
+	}
+}
+
+#if 0
+/*
+* vinagre_spinner_set_timeout:
+* @spinner: a #VinagreSpinner
+* @timeout: time delay between updates to the spinner.
+*
+* Sets the timeout delay for spinner updates.
+**/
+void
+vinagre_spinner_set_timeout (VinagreSpinner *spinner,
+			   guint         timeout)
+{
+	VinagreSpinnerPrivate *priv = spinner->priv;
+
+	if (timeout != priv->timeout)
+	{
+		vinagre_spinner_stop (spinner);
+
+		priv->timeout = timeout;
+
+		if (priv->spinning)
+		{
+			vinagre_spinner_start (spinner);
+		}
+	}
+}
+#endif
+
+static void
+vinagre_spinner_size_request (GtkWidget *widget,
+			   GtkRequisition *requisition)
+{
+	VinagreSpinner *spinner = VINAGRE_SPINNER (widget);
+	VinagreSpinnerPrivate *priv = spinner->priv;
+
+	if ((priv->need_load &&
+	     !vinagre_spinner_load_images (spinner)) ||
+            priv->images == NULL)
+	{
+		requisition->width = requisition->height = 0;
+		gtk_icon_size_lookup_for_settings (gtk_widget_get_settings (widget),
+						   priv->size,
+						   &requisition->width,
+					           &requisition->height);
+		return;
+	}
+
+	requisition->width = priv->images->width;
+	requisition->height = priv->images->height;
+
+	/* FIXME fix this hack */
+	/* allocate some extra margin so we don't butt up against toolbar edges */
+	if (priv->size != GTK_ICON_SIZE_MENU)
+	{
+		requisition->width += 2;
+		requisition->height += 2;
+	}
+}
+
+static void
+vinagre_spinner_map (GtkWidget *widget)
+{
+	VinagreSpinner *spinner = VINAGRE_SPINNER (widget);
+	VinagreSpinnerPrivate *priv = spinner->priv;
+
+	GTK_WIDGET_CLASS (parent_class)->map (widget);
+
+	if (priv->spinning)
+	{
+		vinagre_spinner_start (spinner);
+	}
+}
+
+static void
+vinagre_spinner_unmap (GtkWidget *widget)
+{
+	VinagreSpinner *spinner = VINAGRE_SPINNER (widget);
+
+	vinagre_spinner_remove_update_callback (spinner);
+
+	GTK_WIDGET_CLASS (parent_class)->unmap (widget);
+}
+
+static void
+vinagre_spinner_dispose (GObject *object)
+{
+	VinagreSpinner *spinner = VINAGRE_SPINNER (object);
+
+	g_signal_handlers_disconnect_by_func
+			(spinner->priv->icon_theme,
+			 G_CALLBACK (icon_theme_changed_cb), spinner);
+
+	G_OBJECT_CLASS (parent_class)->dispose (object);
+}
+
+static void
+vinagre_spinner_finalize (GObject *object)
+{
+	VinagreSpinner *spinner = VINAGRE_SPINNER (object);
+
+	vinagre_spinner_remove_update_callback (spinner);
+	vinagre_spinner_unload_images (spinner);
+
+	g_object_unref (spinner->priv->cache);
+
+	G_OBJECT_CLASS (parent_class)->finalize (object);
+}
+
+static void
+vinagre_spinner_screen_changed (GtkWidget *widget,
+			      GdkScreen *old_screen)
+{
+	VinagreSpinner *spinner = VINAGRE_SPINNER (widget);
+	VinagreSpinnerPrivate *priv = spinner->priv;
+	GdkScreen *screen;
+
+	if (GTK_WIDGET_CLASS (parent_class)->screen_changed)
+	{
+		GTK_WIDGET_CLASS (parent_class)->screen_changed (widget, old_screen);
+	}
+
+	screen = gtk_widget_get_screen (widget);
+
+	/* FIXME: this seems to be happening when then spinner is destroyed!? */
+	if (old_screen == screen)
+		return;
+
+	/* We'll get mapped again on the new screen, but not unmapped from
+	 * the old screen, so remove timeout here.
+	 */
+	vinagre_spinner_remove_update_callback (spinner);
+
+	vinagre_spinner_unload_images (spinner);
+
+	if (old_screen != NULL)
+	{
+		g_signal_handlers_disconnect_by_func
+			(gtk_icon_theme_get_for_screen (old_screen),
+			 G_CALLBACK (icon_theme_changed_cb), spinner);
+	}
+
+	priv->icon_theme = gtk_icon_theme_get_for_screen (screen);
+	g_signal_connect (priv->icon_theme, "changed",
+			  G_CALLBACK (icon_theme_changed_cb), spinner);
+}
+
+static void
+vinagre_spinner_class_init (VinagreSpinnerClass *class)
+{
+	GObjectClass *object_class =  G_OBJECT_CLASS (class);
+	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
+
+	parent_class = g_type_class_peek_parent (class);
+
+	object_class->dispose = vinagre_spinner_dispose;
+	object_class->finalize = vinagre_spinner_finalize;
+
+	widget_class->expose_event = vinagre_spinner_expose;
+	widget_class->size_request = vinagre_spinner_size_request;
+	widget_class->map = vinagre_spinner_map;
+	widget_class->unmap = vinagre_spinner_unmap;
+	widget_class->screen_changed = vinagre_spinner_screen_changed;
+
+	g_type_class_add_private (object_class, sizeof (VinagreSpinnerPrivate));
+}
+
+/*
+ * vinagre_spinner_new:
+ *
+ * Create a new #VinagreSpinner. The spinner is a widget
+ * that gives the user feedback about network status with
+ * an animated image.
+ *
+ * Return Value: the spinner #GtkWidget
+ **/
+GtkWidget *
+vinagre_spinner_new (void)
+{
+	return GTK_WIDGET (g_object_new (VINAGRE_TYPE_SPINNER, NULL));
+}
diff --git a/vinagre/vinagre-spinner.h b/vinagre/vinagre-spinner.h
new file mode 100644
index 0000000..03d58dd
--- /dev/null
+++ b/vinagre/vinagre-spinner.h
@@ -0,0 +1,89 @@
+/*
+ * vinagre-spinner.h
+ * This file is part of vinagre
+ *
+ * Copyright (C) 2005 - Paolo Maggi 
+ * Copyright (C) 2000 - Eazel, Inc. 
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, 
+ * Boston, MA 02111-1307, USA.
+ */
+ 
+/*
+ * This widget was originally written by Andy Hertzfeld <andy eazel com> for
+ * Nautilus.
+ */
+
+#ifndef __VINAGRE_SPINNER_H__
+#define __VINAGRE_SPINNER_H__
+
+#include <gtk/gtk.h>
+
+G_BEGIN_DECLS
+
+/*
+ * Type checking and casting macros
+ */
+#define VINAGRE_TYPE_SPINNER		(vinagre_spinner_get_type ())
+#define VINAGRE_SPINNER(o)		(G_TYPE_CHECK_INSTANCE_CAST ((o), VINAGRE_TYPE_SPINNER, VinagreSpinner))
+#define VINAGRE_SPINNER_CLASS(k)		(G_TYPE_CHECK_CLASS_CAST((k), VINAGRE_TYPE_SPINNER, VinagreSpinnerClass))
+#define VINAGRE_IS_SPINNER(o)		(G_TYPE_CHECK_INSTANCE_TYPE ((o), VINAGRE_TYPE_SPINNER))
+#define VINAGRE_IS_SPINNER_CLASS(k)	(G_TYPE_CHECK_CLASS_TYPE ((k), VINAGRE_TYPE_SPINNER))
+#define VINAGRE_SPINNER_GET_CLASS(o)	(G_TYPE_INSTANCE_GET_CLASS ((o), VINAGRE_TYPE_SPINNER, VinagreSpinnerClass))
+
+
+/* Private structure type */
+typedef struct _VinagreSpinnerPrivate	VinagreSpinnerPrivate;
+
+/*
+ * Main object structure
+ */
+typedef struct _VinagreSpinner		VinagreSpinner;
+
+struct _VinagreSpinner
+{
+	GtkWidget parent;
+
+	/*< private >*/
+	VinagreSpinnerPrivate *priv;
+};
+
+/*
+ * Class definition
+ */
+typedef struct _VinagreSpinnerClass	VinagreSpinnerClass;
+
+struct _VinagreSpinnerClass
+{
+	GtkWidgetClass parent_class;
+};
+
+/*
+ * Public methods
+ */
+GType		vinagre_spinner_get_type	(void) G_GNUC_CONST;
+
+GtkWidget      *vinagre_spinner_new	(void);
+
+void		vinagre_spinner_start	(VinagreSpinner *throbber);
+
+void		vinagre_spinner_stop	(VinagreSpinner *throbber);
+
+void		vinagre_spinner_set_size	(VinagreSpinner *spinner,
+					 GtkIconSize   size);
+
+G_END_DECLS
+
+#endif /* __VINAGRE_SPINNER_H__ */



[Date Prev][Date Next]   [Thread Prev][Thread Next]   [Thread Index] [Date Index] [Author Index]
diff --git a/vinagre/vinagre-notebook.c b/vinagre/vinagre-notebook.c
index 33797fa..374561e 100644
--- a/vinagre/vinagre-notebook.c
+++ b/vinagre/vinagre-notebook.c
@@ -28,6 +28,7 @@
 #include "vinagre-utils.h"
 #include "vinagre-dnd.h"
 #include "vinagre-prefs.h"
+#include "vinagre-spinner.h"
 
 #define VINAGRE_NOTEBOOK_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), VINAGRE_TYPE_NOTEBOOK, VinagreNotebookPrivate))
 
@@ -196,6 +197,27 @@ vinagre_notebook_update_ui_sentitivity (VinagreNotebook *nb)
   active = (nb->priv->active_tab) &&
 	   (vinagre_tab_get_state (VINAGRE_TAB (nb->priv->active_tab)) == VINAGRE_TAB_STATE_CONNECTED);
   gtk_action_group_set_sensitive (action_group, active);
+
+  if (nb->priv->active_tab)
+    {
+      GtkWidget *spinner, *icon;
+
+      spinner = g_object_get_data (G_OBJECT (nb->priv->active_tab), "spinner");
+      icon = g_object_get_data (G_OBJECT (nb->priv->active_tab), "icon");
+
+      if (vinagre_tab_get_state (VINAGRE_TAB (nb->priv->active_tab)) == VINAGRE_TAB_STATE_CONNECTED)
+	{
+	  gtk_widget_hide (spinner);
+	  vinagre_spinner_stop (VINAGRE_SPINNER (spinner));
+	  gtk_widget_show (icon);
+	}
+      else
+	{
+	  gtk_widget_hide (icon);
+	  gtk_widget_show (spinner);
+	  vinagre_spinner_start (VINAGRE_SPINNER (spinner));
+	}
+    }
 }
 
 static void
@@ -467,7 +489,7 @@ build_tab_label (VinagreNotebook *nb,
 {
   GtkWidget *hbox, *label_hbox, *label_ebox;
   GtkWidget *label, *dummy_label;
-  GtkWidget *close_button;
+  GtkWidget *close_button, *spinner;
   GtkRcStyle *rcstyle;
   GtkWidget *image;
   GtkWidget *icon;
@@ -512,7 +534,12 @@ build_tab_label (VinagreNotebook *nb,
 		    G_CALLBACK (close_button_clicked_cb),
 		    tab);
 
-  /* setup site icon, empty by default */
+  /* setup spinner */
+  spinner = vinagre_spinner_new ();
+  vinagre_spinner_set_size (VINAGRE_SPINNER (spinner), GTK_ICON_SIZE_MENU);
+  gtk_box_pack_start (GTK_BOX (label_hbox), spinner, FALSE, FALSE, 0);
+
+  /* setup site icon */
   icon = gtk_image_new_from_icon_name (vinagre_tab_get_icon_name (tab),
 				       GTK_ICON_SIZE_MENU);
   gtk_box_pack_start (GTK_BOX (label_hbox), icon, FALSE, FALSE, 0);
@@ -535,13 +562,13 @@ build_tab_label (VinagreNotebook *nb,
   gtk_widget_show (dummy_label);	
   gtk_widget_show (image);
   gtk_widget_show (close_button);
-  gtk_widget_show (icon);
   
   g_object_set_data (G_OBJECT (hbox), "label", label);
   g_object_set_data (G_OBJECT (tab),  "label", label);
   g_object_set_data (G_OBJECT (hbox), "label-ebox", label_ebox);
   g_object_set_data (G_OBJECT (tab),  "label-ebox", label_ebox);
-  g_object_set_data (G_OBJECT (hbox), "icon", icon);
+  g_object_set_data (G_OBJECT (tab), "spinner", spinner);
+  g_object_set_data (G_OBJECT (tab), "icon", icon);
   g_object_set_data (G_OBJECT (hbox), "close-button", close_button);
   g_object_set_data (G_OBJECT (tab),  "close-button", close_button);
 